from pprint import pprint
from typing import Sequence
import pytest
from sklearn.utils._tags import _safe_tags

from bipartite_learn.model_selection import multipartite_cross_validate
from bipartite_learn.tree import BipartiteExtraTreeRegressor
from bipartite_learn.ensemble import BipartiteExtraTreesRegressor
import bipartite_learn.ensemble

from .utils.test_utils import stopwatch, parse_args
from .utils.make_examples import make_interaction_regression


# Default test params
DEF_PARAMS = dict(
    n_samples=(50, 40),
    n_features=(10, 9),
    noise=0.1,
)

CV_DEF_PARAMS = dict(
    groups=None,
    scoring="average_precision",
    cv=3,
    n_jobs=None,
    verbose=10,
    fit_params=None,
    pre_dispatch="2*n_jobs",
    return_train_score=False,
    error_score='raise',
    diagonal=False,
    train_test_combinations=None,
    pairwise=False,
)


@pytest.fixture(params=[6, .1])
def msl(request):  # min_samples_leaf parameter
    return request.param


@pytest.fixture(params=range(10))
def random_state(request):
    return request.param


def _test_cv(estimator, random_state=None, cv_params=None, **PARAMS):
    PARAMS = DEF_PARAMS | PARAMS
    PARAMS['noise'] = 0.
    pprint(PARAMS)
    cv_params = CV_DEF_PARAMS | (cv_params or {})
    cv_params |= dict(return_estimator=True)
    pprint(cv_params)

    is_pairwise = cv_params["pairwise"] or _safe_tags(estimator, "pairwise")
    cv = cv_params["cv"]
    n_dim = len(PARAMS['n_samples'])

    if isinstance(cv, Sequence):
        if not isinstance(cv[0], int):
            raise TypeError("cv_params['cv'] must be (sequence of) int")

    elif not isinstance(cv, int):
        raise TypeError("cv_params['cv'] must be (sequence of) int")

    else:  # isinstance(cv_params["cv"], int)
        cv = (cv, cv)

    if is_pairwise:
        if PARAMS["n_features"] != PARAMS["n_samples"]:
            raise ValueError(
                "Pairwise estimators must receive square Xs satisfying"
                f"attrs == n_samples (PARAMS['n_features'] != PARAMS['n_features'])"
            )
        n_samples = PARAMS["n_samples"]
        # For pairwise data, each fold must be square
        # Sum Xi.n_samples where Xi is a training fold of X
        fold_n_samples = sum(s-(s//cvi) for s, cvi in zip(n_samples, cv))
    else:
        # Sum Yi.n_samples where Yi is a training fold of Y
        n_features = sum(PARAMS["n_features"])

    with stopwatch('Generating data...'):
        X, y = make_interaction_regression(random_state=random_state, **PARAMS)
        y = y > y.mean()
    
    # TODO: random_state
    cv_res = multipartite_cross_validate(estimator, X=X, y=y, **cv_params)
    pprint(cv_res)

    for estimator in cv_res["estimator"]:
        if is_pairwise:
            # When cv does not divide n_samples perfectly, it distributes the
            # remainders across folds (see np.array_split), so that each fold
            # size can vary on one unit. Naturally, this happens for the n_dim
            # X matrices, so n_features_in_ can be up to n_dim smaller.
            assert (estimator.n_features_in_ - fold_n_samples) <= n_dim
        else:
            assert estimator.n_features_in_ == n_features
    
    return cv_res


def test_cv_tree(msl, random_state, **PARAMS):
    PARAMS = DEF_PARAMS | PARAMS
    return _test_cv(
        estimator=BipartiteExtraTreeRegressor(
            min_samples_leaf=msl,
            random_state=random_state,
        ),
        random_state=random_state,
        **PARAMS,
    )


def test_cv_ensemble(msl, random_state, **PARAMS):
    PARAMS = DEF_PARAMS | PARAMS
    return _test_cv(
        estimator=BipartiteExtraTreesRegressor(
            n_estimators=10,
            min_samples_leaf=msl,
            random_state=random_state,
        ),
        random_state=random_state,
        **PARAMS,
    )


def test_cv_semisupervised_ensemble(msl, random_state, **PARAMS):
    PARAMS = DEF_PARAMS | PARAMS
    return _test_cv(
        estimator=bipartite_learn.ensemble.BipartiteExtraTreesRegressorSS(
            n_estimators=10,
            min_samples_leaf=msl,
            random_state=random_state,
            bipartite_adapter='gmosa',
            criterion='squared_error',
        ),
        random_state=random_state,
        **PARAMS,
    )


def test_cv_pairwise_tag(msl, random_state, **PARAMS):
    PARAMS = DEF_PARAMS | PARAMS
    PARAMS["n_features"] = PARAMS["n_samples"]  # Make Xs square
    cv_params = dict(pairwise=False)  # tag will be gotten from the estimator

    estimator = BipartiteExtraTreeRegressor(
        min_samples_leaf=msl,
        random_state=random_state,
    )

    # Monkey-patch pairwise tag
    old_tags = _safe_tags(estimator)
    estimator.__class__._more_tags = (
        lambda self: old_tags | dict(pairwise=True)
    )
    # Revert patch even if an exception is raised
    try:
        cv_res = _test_cv(
            estimator,
            random_state=random_state,
            cv_params=cv_params,
            **PARAMS,
        )
    finally:
        estimator.__class__._more_tags = lambda self: old_tags

    assert not _safe_tags(estimator, "pairwise")
    return cv_res


def test_cv_pairwise_parameter(msl, random_state, **PARAMS):
    PARAMS = DEF_PARAMS | PARAMS
    cv_params = dict(pairwise=True)
    estimator = BipartiteExtraTreeRegressor(
        min_samples_leaf=msl,
        random_state=random_state,
    )
    PARAMS["n_features"] = PARAMS["n_samples"]  # Make Xs square

    cv_res = _test_cv(
        estimator, cv_params=cv_params, random_state=random_state, **PARAMS,
    )

    assert not _safe_tags(estimator, "pairwise")
    return cv_res


import logging
import warnings
import pytest
from sklearn.utils.validation import check_random_state
from sklearn.dummy import DummyRegressor, DummyClassifier
from sklearn.utils._testing import assert_allclose
from sklearn.metrics import mean_absolute_error
from sklearn.metrics.pairwise import rbf_kernel
from bipartite_learn.base import BaseMultipartiteEstimator
from bipartite_learn.utils import all_estimators
from bipartite_learn.melter import row_cartesian_product
from .utils.make_examples import (
    make_interaction_blobs,
    make_interaction_regression,
)


def apply_rbf_kernel(regression_data):
    X, Y, x, y = regression_data
    X = [rbf_kernel(X[0]), rbf_kernel(X[1])]
    x = row_cartesian_product(X)
    return X, Y, x, y


@pytest.fixture(params=range(3))
def random_state(request):
    return check_random_state(request.param)


@pytest.fixture
def n_samples(request):
    return (11, 13)


@pytest.fixture
def n_features(request):
    return (5, 7)


@pytest.fixture
def classification_data(n_samples, n_features, random_state):
    X, Y, x, y = make_interaction_blobs(
        return_molten=True,
        n_features=n_features,
        n_samples=n_samples,
        random_state=random_state,
        noise=0.0,
        centers=10,
        row_kwargs={'center_box': [.3, .7], 'cluster_std': .1},
        col_kwargs={'center_box': [.3, .7], 'cluster_std': .1},
    )
    return X, Y, x, y


@pytest.fixture
def regression_data(n_samples, n_features, random_state):
    X, Y, x, y = make_interaction_regression(
        n_samples=n_samples,
        n_features=n_features,
        n_targets=None,
        min_target=0.0,
        max_target=1.0,
        noise=0.0,
        return_molten=True,
        return_tree=False,
        random_state=random_state,
        max_depth=None,
    )
    return X, Y, x, y


@pytest.fixture
def pairwise_regression_data(regression_data):
    return apply_rbf_kernel(regression_data)


@pytest.fixture
def pairwise_classification_data(classification_data):
    return apply_rbf_kernel(classification_data)


@pytest.mark.parametrize(
    'estimator_name, estimator',
    (
        e for e in all_estimators("regressor")
        if not issubclass(e[1], BaseMultipartiteEstimator)
    )
)
def test_all_monopartite_regressors(
    regression_data,
    pairwise_regression_data,
    estimator_name,
    estimator,
):
    estimator = estimator()

    if estimator._get_tags()['pairwise']:
        X, Y, x, y = pairwise_regression_data
    else:
        X, Y, x, y = regression_data

    estimator.fit(X[0], Y)
    score = estimator.score(X[0], Y)

    logging.info(f"{estimator_name}'s score: {score}")
    assert score > 0.1


@pytest.mark.parametrize(
    'estimator_name, estimator', all_estimators("regressor", "multipartite")
)
def test_all_multipartite_regressors(
    regression_data,
    pairwise_regression_data,
    estimator_name,
    estimator,
):
    estimator = estimator()

    if estimator._get_tags()['pairwise']:
        X, Y, x, y = pairwise_regression_data
    else:
        X, Y, x, y = regression_data

    estimator.fit(X, Y)

    # The two formats must be supported:
    pred1 = estimator.predict(X)
    try:
        pred2 = estimator.predict(x)
        assert_allclose(pred1, pred2)
    except ValueError:
        warnings.warn(f"{estimator_name} does not accept melted input.")

    score = estimator.score(X, y)

    logging.info(f"{estimator_name}'s score: {score}")
    assert score > 0.1


@pytest.mark.parametrize(
    'estimator_name, estimator', all_estimators("classifier", "multipartite")
)
def test_all_multipartite_classifiers(
    classification_data,
    pairwise_classification_data,
    estimator_name,
    estimator,
):
    estimator = estimator()
    dummy = DummyClassifier()

    if estimator._get_tags()['pairwise']:
        X, Y, x, y = pairwise_classification_data
    else:
        X, Y, x, y = classification_data

    estimator.fit(X, Y)
    dummy.fit(x, y)

    # The two formats must be supported:
    pred1 = estimator.predict(X)
    try:
        pred2 = estimator.predict(x)
        assert_allclose(pred1, pred2)
    except ValueError:
        warnings.warn(f"{estimator_name} does not accept melted input.")

    score = estimator.score(X, y)

    logging.info(f"{estimator_name}'s score: {score}")
    assert score > dummy.score(x, y)



from bipartite_learn.model_selection._validation import (
    multipartite_cross_validate,
)
from bipartite_learn.model_selection._split import (
    check_multipartite_cv,
    CrossValidatorNDWrapper,
    make_train_test_splitter_nd,
    make_kfold_nd,
    train_test_split_nd,
)
from bipartite_learn.model_selection._search import MultipartiteGridSearchCV
from bipartite_learn.model_selection._search import MultipartiteRandomizedSearchCV


__all__ = [
    "split_train_test_nd",
    "make_train_test_splitter_nd",
    "multipartite_cross_validate",
    "CrossValidatorNDWrapper",
    "MultipartiteGridSearchCV",
    "MultipartiteRandomizedSearchCV",
]



# TODO: rename
import numpy as np
from sklearn.neighbors import KNeighborsRegressor
from sklearn.linear_model import Ridge
from sklearn.utils._testing import assert_allclose
from sklearn.utils.validation import check_symmetric
from imblearn.pipeline import make_pipeline

from bipartite_learn.matrix_factorization._nrlmf import NRLMF
from bipartite_learn.matrix_factorization._dnilmf import DNILMF
from bipartite_learn.preprocessing.multipartite import DTHybridSampler
from bipartite_learn.preprocessing.monopartite import TargetKernelLinearCombiner
from bipartite_learn.wrappers import (
    BipartiteLocalWrapper, MultipartiteSamplerWrapper,
)
from .utils.test_utils import gen_mock_data, DEF_PARAMS, parse_args


def test_nrlmf(**params):
    params = DEF_PARAMS | params
    params['noise'] = 0
    params["shape"] = params["nattrs"]  # X must be kernel matrices

    XX, Y, = gen_mock_data(**params, melt=False)

    nrlmf = NRLMF(verbose=True)
    XXt, Yt = nrlmf.fit_resample(XX, Y)
    nrlmf.predict([XX[0][:5], XX[1][:3]])
    Y_positive = Y.astype(bool)

    assert Yt[Y_positive].mean() > Yt[~Y_positive].mean()

    return XXt


def test_nrlmf_fit_predict(**params):
    params = DEF_PARAMS | params
    params['noise'] = 0
    params["shape"] = params["nattrs"]  # X must be kernel matrices

    XX, Y, = gen_mock_data(**params, melt=False)
    XX = [check_symmetric(Xi, raise_warning=False) for Xi in XX]
    np.fill_diagonal(XX[0], 1.)
    np.fill_diagonal(XX[1], 1.)

    nrlmf = NRLMF(verbose=True, random_state=0)

    pred1 = nrlmf.fit(XX, Y).predict(XX)
    U1, V1 = nrlmf.U, nrlmf.V

    pred2 = nrlmf.fit_predict(XX, Y)
    U2, V2 = nrlmf.U, nrlmf.V

    assert_allclose(U1, U2)
    assert_allclose(V1, V2)
    assert_allclose(pred1, pred2)


def test_dnilmf_fit_predict(**params):
    params = DEF_PARAMS | params
    params['noise'] = 0
    params["shape"] = params["nattrs"]  # X must be kernel matrices

    XX, Y, = gen_mock_data(**params, melt=False)
    XX = [check_symmetric(Xi, raise_warning=False) for Xi in XX]
    np.fill_diagonal(XX[0], 1.)
    np.fill_diagonal(XX[1], 1.)

    dnilmf = DNILMF(verbose=True, random_state=0)
    pred1 = dnilmf.fit(XX, Y).predict(XX)
    U1, V1 = dnilmf.U, dnilmf.V

    pred2 = dnilmf.fit_predict(XX, Y)
    U2, V2 = dnilmf.U, dnilmf.V

    assert_allclose(U1, U2)
    assert_allclose(V1, V2)
    assert_allclose(pred1, pred2)


def test_dnilmf(**params):
    params = DEF_PARAMS | params
    params['noise'] = 0
    params["shape"] = params["nattrs"]  # X must be kernel matrices

    XX, Y, = gen_mock_data(**params, melt=False)

    dnilmf = DNILMF(verbose=True)
    XXt, Yt = dnilmf.fit_resample(XX, Y)
    dnilmf.predict([XX[0][:5], XX[1][:3]])
    Y_positive = Y.astype(bool)

    assert Yt[Y_positive].mean() > Yt[~Y_positive].mean()

    return XXt


def test_dthybrid(**params):
    params = DEF_PARAMS | params
    params["noise"] = 0
    params["shape"] = params["nattrs"]  # X must be kernel matrices
    rng = np.random.default_rng(params["seed"])

    shape, nattrs = params["shape"], params["nattrs"]
    XX = [
        rng.random((shape[0], nattrs[0])),
        rng.random((shape[1], nattrs[1])),
    ]
    Y = rng.choice((0., 1.), size=shape)

    dthybrid = DTHybridSampler()
    XXt, Yt = dthybrid.fit_resample(XX, Y)
    Y_positive = Y.astype(bool)

    assert Yt[Y_positive].mean() > Yt[~Y_positive].mean()


def test_blmnii(**params):
    params = DEF_PARAMS | params
    params["noise"] = 0
    params["shape"] = params["nattrs"]  # X must be kernel matrices

    # Gaussian interaction profile
    gip_transformer = MultipartiteSamplerWrapper(TargetKernelLinearCombiner())

    XX, Y, = gen_mock_data(**params, melt=False)

    blmnii = make_pipeline(
        gip_transformer,
        BipartiteLocalWrapper(
            primary_estimator=KNeighborsRegressor(),
            secondary_estimator=Ridge(),
        ),
    )

    Yt = blmnii.fit(XX, Y).predict(XX).reshape(Y.shape)
    Y_positive = Y.astype(bool)

    assert Yt[Y_positive].mean() > Yt[~Y_positive].mean()


def main(**params):
    test_nrlmf(**params)
    test_nrlmf_fit_predict(**params)
    test_dnilmf(**params)
    test_dnilmf_fit_predict(**params)
    test_dthybrid(**params)
    test_blmnii(**params)


if __name__ == "__main__":
    args = parse_args()
    main(**vars(args))


"""Set of tools to apply standard estimators to bipartite datasets.

TODO: Docs.
TODO: check fit inputs.
"""
from __future__ import annotations
from typing import Callable, Sequence
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin, MetaEstimatorMixin, clone
from sklearn.utils.metaestimators import available_if
from sklearn.utils._tags import _safe_tags
from sklearn.utils.validation import check_is_fitted
from sklearn.exceptions import NotFittedError
from sklearn.utils import check_random_state
from .melter import melt_multipartite_dataset
from .utils import check_multipartite_params
from .base import BaseMultipartiteEstimator, BaseMultipartiteSampler
from .utils import _X_is_multipartite


def _estimator_has(attr):
    """Check that primary estimators has `attr`.

    Used together with `avaliable_if` in `GlobalSingleOutputWrapper`.
    """
    def check(self):
        # raise original `AttributeError` if `attr` does not exist
        getattr(self.estimator, attr)
        return True

    return check
     

def _secondary_estimators_have(attr):
    """Check that primary estimators has `attr`.

    Used together with `avaliable_if` in `BipartiteLocalWrapper`.
    """
    def check(self):
        # raise original `AttributeError` if `attr` does not exist
        getattr(self.secondary_rows_estimator_, attr)
        getattr(self.secondary_cols_estimator_, attr)
        return True

    return check


def _primary_estimators_have(attr):
    """Check that secondary estimators have `attr`.

    Used together with `avaliable_if` in `BipartiteLocalWrapper`.
    """
    def check(self):
        # raise original `AttributeError` if `attr` does not exist
        getattr(self.primary_rows_estimator_, attr)
        getattr(self.primary_cols_estimator_, attr)
        return True

    return check


class IncompatibleEstimatorsError(ValueError):
    """Raised when user tries to wrap incompatible estimators."""


class GlobalSingleOutputWrapper(BaseMultipartiteEstimator, MetaEstimatorMixin):
    def __init__(
        self,
        estimator: BaseEstimator,
        random_state: int | np.random.RandomState | None = None,
        subsample_negatives: bool = False,
    ):
        """Wraps a standard estimator/transformer to work on PU n-partite data.

        With n-partite interaction data, X is actualy a list of sample
        attribute matrices, one for each group of samples (drugs and targets,
        for instance). y is then an interaction matrix (or tensor) denoting
        interaction between the instances of each group.

        We are currently assuming positive-unlabeled data, in which y == 1
        denotes positive and known interaction and y == 0 denotes unknown
        (positive or negative).

        This wrapper melts the n-partite X/y data before passing them to the
        estimator, so that monopartite estimators (standard ones) can be used
        with n-partite cross-validators and pipelines.

        Since all multipartite sets are considered jointly, not consecutively,
        and only single output estimators are compatible, this adaptation
        format is called Global Single Output (GSO) [Pliakos _et al._, 2020].

        Parameters
        ----------

        estimator : sklearn estimator or transformer
            Estimator/transformed to receive processed data.

        random_state : int or np.random.RandomState or None
            Seed or random state object to be used for negative data
            subsampling. If None, it takes its value from self.estimator.
            Has no effect if subsample_negatives == False.

        subsample_negatives : boolean
            Indicates wether to use all unknown data available (False) or to
            randomly subsample negative pairs (assumed zero-labeled, y == 0)
            to have a balanced dataset.
        """
        self.estimator = estimator
        self.random_state = random_state
        self.subsample_negatives = subsample_negatives

    def _melt_input(self, X, y=None):
        if not _X_is_multipartite(X):
            return X, y

        return melt_multipartite_dataset(
            X,
            y,
            subsample_negatives=self.subsample_negatives,
            random_state=self.random_state_,
        )

    def fit(self, X, y=None, **fit_params):
        self.random_state_ = check_random_state(self.random_state)
        Xt, yt = self._melt_input(X, y=y)
        self.estimator = clone(self.estimator)
        self.estimator.fit(X=Xt, y=yt, **fit_params)
        return self

    @available_if(_estimator_has("predict"))
    def predict(self, X):
        X, _ = self._melt_input(X)
        return self.estimator.predict(X)

    @available_if(_estimator_has("fit_predict"))
    def fit_predict(self, X, y=None, **fit_params):
        Xt, yt = self._melt_input(X, y=y)
        return self.estimator.fit_predict(Xt, yt, **fit_params)

    @available_if(_estimator_has("predict_proba"))
    def predict_proba(self, X, **predict_proba_params):
        X, _ = self._melt_input(X)
        return self.estimator.predict_proba(X, **predict_proba_params)

    @available_if(_estimator_has("decision_function"))
    def decision_function(self, X):
        X, _ = self._melt_input(X)
        return self.estimator.decision_function(X)

    @available_if(_estimator_has("score_samples"))
    def score_samples(self, X):
        X, _ = self._melt_input(X)
        return self.estimator.score_samples(X)

    @available_if(_estimator_has("predict_log_proba"))
    def predict_log_proba(self, X, **predict_log_proba_params):
        X, _ = self._melt_input(X)
        return self.estimator.predict_log_proba(X, **predict_log_proba_params)

    @available_if(_estimator_has("transform"))
    def transform(self, X):
        Xt, _ = self._melt_input(X)
        return self.transform(Xt)

    @available_if(_estimator_has("fit_transform"))
    def fit_transform(self, X, y=None, **fit_params):
        Xt, yt = self._melt_input(X, y=y)
        return self.estimator.fit_transform(Xt, yt, **fit_params)
        
    @available_if(_estimator_has("inverse_transform"))
    def inverse_transform(self, Xt):
        X, _ = self._melt_input(Xt)
        return self.estimator.inverse_transform(X)

    @available_if(_estimator_has("score"))
    def score(self, X, y=None, sample_weight=None):
        X, y = self._melt_input(X, y=y)
        score_params = {}
        if sample_weight is not None:
            score_params["sample_weight"] = sample_weight
        return self.estimator.score(X, y, **score_params)

    @property
    def _estimator_type(self):
        return getattr(self.estimator, "_estimator_type", None)

    @property
    def classes_(self):
        """The classes labels. Only exist if the estimator is a classifier."""
        return self.estimator.classes_

    @property
    def n_features_in_(self):
        """Number of features seen during `fit`."""
        return self.estimator.n_features_in_

    @property
    def feature_names_in_(self):
        """Names of features seen during `fit`."""
        return self.estimator.feature_names_in_

    def __sklearn_is_fitted__(self):
        """Indicate whether the estimator has been fit."""
        try:
            check_is_fitted(self.estimator)
            return True
        except NotFittedError:
            return False


# TODO: docs.
class BipartiteLocalWrapper(BaseMultipartiteEstimator):
    def __init__(
        self,
        primary_estimator: BaseEstimator | Sequence[BaseEstimator],
        secondary_estimator: BaseEstimator | Sequence[BaseEstimator],
        combine_predictions_func: str | Callable = np.mean,
        combine_func_kwargs: dict | None = None,
        independent_labels: bool = True,
    ):
        """Wraps a standard estimator/transformer to work on PU n-partite data.

        With n-partite interaction data, X is actualy a list of sample
        attribute matrices, one for each group of samples (drugs and targets,
        for instance). y is then an interaction matrix (or tensor) denoting
        interaction between the instances of each group.

        We are currently assuming positive-unlabeled data, in which y == 1
        denotes positive and known interaction and y == 0 denotes unknown
        (positive or negative).

        This wrapper melts the n-partite X/y data before passing them to the
        estimator, so that monopartite estimators (standard ones) can be used
        with n-partite cross-validators and pipelines.

        Parameters
        ----------

        estimator : sklearn estimator or transformer
            Estimator/transformed to receive processed data.

        random_state : int or np.random.RandomState or None
            Seed or random state object to be used for negative data
            subsampling. If None, it takes its value from self.estimator.
            Has no effect if subsample_negatives == False.

        subsample_negatives : boolean
            Indicates wether to use all unknown data available (False) or to
            randomly subsample negative pairs (assumed zero-labeled, y == 0)
            to have a balanced dataset.
        """
        self.primary_estimator = primary_estimator
        self.secondary_estimator = secondary_estimator
        self.independent_labels = independent_labels
        self.combine_predictions_func = combine_predictions_func
        self.combine_func_kwargs = combine_func_kwargs

        # FIXME: no validation in init, but some properties depend on this.
        # FIXME: messes up check_is_fitted
        self._check_estimators()  
    
    def _check_estimators(self):
        primary_estimators, secondary_estimators = check_multipartite_params(
            self.primary_estimator, self.secondary_estimator
        )

        for estimator in (*primary_estimators, *secondary_estimators):
            # TODO: 'multioutput_only' tag should imply 'multioutput' tag
            if not (
                _safe_tags(estimator, "multioutput") or
                _safe_tags(estimator, "multioutput_only")
            ):
                raise IncompatibleEstimatorsError(
                    f"All estimators wrapped by {self.__class__.__name__} "
                    f"must support multioutput functionality but {estimator} "
                    "does not. Some meta-estimators defined in scikit-learn "
                    "may be useful, check https://scikit-learn.org/stable/"
                    "modules/multiclass.html"
                )

        self.primary_rows_estimator_ = clone(primary_estimators[0])
        self.primary_cols_estimator_ = clone(primary_estimators[1])
        self.secondary_rows_estimator_ = clone(secondary_estimators[0])
        self.secondary_cols_estimator_ = clone(secondary_estimators[1])

        if (
            getattr(self.secondary_rows_estimator_, "_estimator_type") !=
            getattr(self.secondary_cols_estimator_, "_estimator_type")
        ):
            raise IncompatibleEstimatorsError(
                "Secondary estimators must be of the same type (regressor"
                ", classifier, etc.). See https://scikit-learn.org/stable/"
                "developers/develop.html#estimator-types"
            )

        if (
            _safe_tags(self.primary_rows_estimator_, "pairwise") !=
            _safe_tags(self.primary_cols_estimator_, "pairwise")
        ):
            raise IncompatibleEstimatorsError(
                "Both or none of the primary estimators must be pairwise."
            )

    # TODO: organize 'fit_params'
    def fit(self, X, y=None, **fit_params):
        # TODO: check input
        # self._check_estimators()  # FIXME: remove from init.
        self.X_fit_ = X

        if not self.independent_labels:
            self.y_fit_ = y

        self.primary_rows_estimator_.fit(X=X[0], y=y, **fit_params)
        self.primary_cols_estimator_.fit(X=X[1], y=y.T, **fit_params)

        return self

    def _secondary_fit(self, X):
        """Fit secondary estimators at prediction time."""
        check_is_fitted(self)

        # Transposed because they will be used as training labels.
        new_y_rows = self.primary_rows_estimator_.predict(X[0]).T
        new_y_cols = self.primary_cols_estimator_.predict(X[1]).T

        # If the secondary estimators are able to take advantage of correlated
        # labels, we train them with all the labels we have, even if most pre-
        # dicted columns will be discarded. 
        if not self.independent_labels:
            new_y_cols = np.hstack((new_y_cols, self.y_fit_))
            new_y_rows = np.hstack((new_y_rows, self.y_fit_.T))

        self.secondary_rows_estimator_.fit(self.X_fit_[0], new_y_cols)
        self.secondary_cols_estimator_.fit(self.X_fit_[1], new_y_rows)

        return self
    
    def _combine_predictions(self, rows_pred, cols_pred):
        return np.apply_along_axis(
            func1d=self.combine_predictions_func,
            axis=0,
            arr=(rows_pred, cols_pred),
            **(self.combine_func_kwargs or {}),
        ).reshape(-1)

    @available_if(_secondary_estimators_have("predict"))
    def predict(self, X):
        self._secondary_fit(X)

        rows_pred = self.secondary_rows_estimator_.predict(X[0])
        cols_pred = self.secondary_cols_estimator_.predict(X[1]).T

        if not self.independent_labels:
            # Get only the predictions corresponding to the instances being
            # predicted.
            rows_pred = rows_pred[:, :X[1].shape[0]]
            cols_pred = cols_pred[:X[0].shape[0]]

        return self._combine_predictions(rows_pred, cols_pred)
    
    @available_if(lambda self: _primary_estimators_have("fit_predict") or
                               _secondary_estimators_have("fit_predict"))
    def fit_predict(self, X, y=None, **fit_params):
        # TODO: check input
        # self._check_estimators()  # FIXME: remove from init.
        self.X_fit_ = X
        self.primary_rows_estimator_ = clone(self.estimator_rows)
        self.primary_cols_estimator_ = clone(self.estimator_cols)
        self.secondary_rows_estimator_ = clone(self.secondary_estimator_rows)
        self.secondary_cols_estimator_ = clone(self.secondary_estimator_cols)

        if not self.independent_labels:
            self.y_fit_ = y
        
        # Transposed because they will be used as training labels.
        if hasattr(self.primary_rows_estimator_, "fit_predict"):
            new_y_rows = self.primary_rows_estimator_.fit_predict(
                X[0], y, **fit_params).T
            new_y_cols = self.primary_cols_estimator_.fit_predict(
                X[1], y.T, **fit_params).T
        else:
            self.fit(X, y, **fit_params)
            new_y_rows = self.primary_rows_estimator_.predict(X[0]).T
            new_y_cols = self.primary_cols_estimator_.predict(X[1]).T

        # If the secondary estimators are able to take advantage of correlated
        # labels, we train them with all the labels we have, even if most pre-
        # dicted columns will be discarded. 
        if not self.independent_labels:
            new_y_cols = np.hstack((new_y_cols, y))
            new_y_rows = np.hstack((new_y_rows, y.T))

        if hasattr(self.secondary_rows_estimator_, "fit_predict"):
            rows_pred = self.secondary_rows_estimator_.fit_predict(
                X[0], new_y_cols, **fit_params)
            cols_pred = self.secondary_cols_estimator_.fit_predict(
                X[1], new_y_rows, **fit_params).T
        else:
            self.secondary_rows_estimator_.fit(X[0], new_y_cols, **fit_params)
            self.secondary_cols_estimator_.fit(X[1], new_y_rows, **fit_params)
            rows_pred = self.secondary_rows_estimator_.predict(X[0])
            cols_pred = self.secondary_cols_estimator_.predict(X[1]).T

        if not self.independent_labels:
            # Get only the predictions corresponding to the instances being
            # predicted.
            rows_pred = rows_pred[:, :X[1].shape[0]]
            cols_pred = cols_pred[:X[0].shape[0]]

        return self._combine_predictions(rows_pred, cols_pred)

    @available_if(_secondary_estimators_have("predict_proba"))
    def predict_proba(self, X, **predict_proba_params):
        self._secondary_fit(X)

        rows_pred = self.secondary_rows_estimator_.predict_proba(
            X[0], **predict_proba_params)
        cols_pred = self.secondary_cols_estimator_.predict_proba(
            X[1], **predict_proba_params).T

        if not self.independent_labels:
            # Get only the predictions corresponding to the instances being
            # predicted.
            rows_pred = rows_pred[:, :X[1].shape[0]]
            cols_pred = cols_pred[:X[0].shape[0]]

        # FIXME: Does not work in some cases, such as 'max()'
        return self._combine_predictions(rows_pred, cols_pred)

    @available_if(_secondary_estimators_have("predict_log_proba"))
    def predict_log_proba(self, X, **predict_log_proba_params):
        self._secondary_fit(X)

        rows_pred = self.secondary_rows_estimator_.predict_log_proba(
            X[0], **predict_log_proba_params)
        cols_pred = self.secondary_cols_estimator_.predict_log_proba(
            X[1], **predict_log_proba_params).T

        if not self.independent_labels:
            # Get only the predictions corresponding to the instances being
            # predicted.
            rows_pred = rows_pred[:, :X[1].shape[0]]
            cols_pred = cols_pred[:X[0].shape[0]]

        # FIXME: Does not work in some cases, such as 'max()'
        return self._combine_predictions(rows_pred, cols_pred)

    @property
    def _estimator_type(self):
        return getattr(self.secondary_rows_estimator_, "_estimator_type", None)

    @property
    def classes_(self):
        """The classes labels. Only exist if the estimator is a classifier."""
        return self.estimator.classes_

    def _more_tags(self):
        # check if the primary estimator expects pairwise input
        return {"pairwise": _safe_tags(self.primary_rows_estimator_, "pairwise")}

    @property
    def n_features_in_(self):
        """Number of features seen during first step `fit` method."""
        return (
            self.primary_rows_estimator_.n_features_in_ +
            self.primary_cols_estimator_.n_features_in_
        )

    @property
    def feature_names_in_(self):
        """Names of features seen during first step `fit` method."""
        return (
            self.primary_rows_estimator_.feature_names_in_ +
            self.primary_cols_estimator_.feature_names_in_
        )

    def __sklearn_is_fitted__(self):
        """Indicate whether the primary estimators have been fit."""
        try:
            check_is_fitted(self.primary_rows_estimator_)
            check_is_fitted(self.primary_cols_estimator_)
            return True
        except NotFittedError:
            # Let the caller check_is_fitted() raise the error
            return False


# FIXME: test get_params()
class MultipartiteTransformerWrapper(BaseMultipartiteEstimator,
                                     TransformerMixin):
    """Manages a transformer for each feature space in multipartite datasets.
    """
    def __init__(
        self,
        transformers: BaseEstimator | Sequence[BaseEstimator],
        ndim: int | None = 2,
    ):
        self.transformers = transformers
        self.ndim = ndim

    def fit(self, X, y=None):
        self._set_transformers()
        self._validate_data(X, y)
        for Xi, yi, transformer in self._roll_axes(X, y):
            transformer.fit(Xi, yi)

    def transform(self, X, y=None):
        self._validate_data(X, y)
        return [
            transformer.transform(Xi)
            for Xi, _, transformer in self._roll_axes(X)
        ]

    def fit_transform(self, X, y=None):
        self._set_transformers()
        self._validate_data(X, y)
        return [
            transformer.fit_transform(Xi, yi)
            for Xi, yi, transformer in self._roll_axes(X, y)
        ]

    def _set_transformers(self):
        """Sets self.transformers_ and self.ndim_ during fit.
        """
        if isinstance(self.transformers, Sequence):
            if self.ndim is None:
                self.ndim_ = len(self.transformers)

            elif self.ndim != len(self.transformers):
                raise ValueError("'self.ndim' must correspond to the number of"
                                 " transformers in self.transformers")
            else:
                self.ndim_ = self.ndim

            self.transformers_ = [clone(t) for t in self.transformers]

        else:  # If a single transformer provided
            if self.ndim is None:
                raise ValueError("'ndim' must be provided if 'transformers' is"
                                 " a single transformer.")
            self.ndim_ = self.ndim
            self.transformers_ = [
                clone(self.transformers) for _ in range(self.ndim_)
            ]

    def _validate_data(self, X, y):
        if len(X) != self.ndim_:
            raise ValueError(f"Wrong dimension. X has {len(X)} items, was exp"
                             "ecting {self.ndim}.")
        # FIXME: validate
        # super()._validate_data(X, y, validate_separately=True)

    def _roll_axes(self, X, y=None):
        if not hasattr(self, "transformers_"):
            raise AttributeError("One must call _set_transformers() before"
                                 "attempting to call _roll_axes()")

        for ax in range(self.ndim_):
            yield X[ax], y, self.transformers_[ax]

            if y is not None:
                y = np.moveaxis(y, 0, -1)

    def __sklearn_is_fitted__(self):
        """Indicate whether the transformers have been fit."""
        if not hasattr(self, "ndim_"):
            return False
        try:
            map(check_is_fitted, self.transformers_)
        except NotFittedError:
            # Let the caller check_is_fitted() raise the error
            return False


# TODO: sampler wrapper and transformer wrapper could share code
class MultipartiteSamplerWrapper(BaseMultipartiteSampler):
    """Manages a sampler for each feature space in multipartite datasets.
    """
    def __init__(
        self,
        samplers: BaseEstimator | Sequence[BaseEstimator],
        ndim: int | None = 2,
    ):
        self.samplers = samplers
        self.ndim = ndim

    def _fit_resample(self, X, y):
        # NOTE: ressampled y is discarded!
        self._set_samplers()
        self._validate_data(X, y)
        return (
            [
                # FIXME: imblearn cannot deal with multi-column y
                # sampler.fit_resample(Xi, yi)[0]  # skip validation
                sampler._fit_resample(Xi, yi)[0]
                for Xi, yi, sampler in self._roll_axes(X, y)
            ],
            y,
        )

    def _set_samplers(self):
        """Sets self.samplers_ and self.ndim_ during fit.
        """
        if isinstance(self.samplers, Sequence):
            if self.ndim is None:
                self.ndim_ = len(self.samplers)

            elif self.ndim != len(self.samplers):
                raise ValueError("'self.ndim' must correspond to the number of"
                                 " samplers in self.samplers")
            else:
                self.ndim_ = self.ndim

            self.samplers_ = [clone(t) for t in self.samplers]

        else:  # If a single sampler provided
            if self.ndim is None:
                raise ValueError("'ndim' must be provided if 'samplers' is"
                                 " a single sampler.")
            self.ndim_ = self.ndim
            self.samplers_ = [
                clone(self.samplers) for _ in range(self.ndim_)
            ]

    def _validate_data(self, X, y):
        if len(X) != self.ndim_:
            raise ValueError(f"Wrong dimension. X has {len(X)} items, was exp"
                             "ecting {self.ndim}.")
        # FIXME: validate
        # super()._validate_data(X, y, validate_separately=True)

    def _roll_axes(self, X, y=None):
        if not hasattr(self, "samplers_"):
            raise AttributeError("One must call _set_samplers() before"
                                 "attempting to call _roll_axes()")

        for ax in range(self.ndim_):
            yield X[ax], y, self.samplers_[ax]

            if y is not None:
                y = np.moveaxis(y, 0, -1)



import numpy as np
from sklearn.datasets import make_regression
from sklearn.utils._testing import assert_allclose
from bipartite_learn.neighbors import WeightedNeighborsRegressor


def test_weighted_neighbors_euclidean():
    X, y = make_regression(n_samples=1000, n_features=500, n_targets=100)
    X.setflags(write=False)

    rng = np.random.default_rng()
    X_test = rng.random((X.shape[0]//2, X.shape[1]))

    wnr = WeightedNeighborsRegressor()
    wnr.fit(X, y)

    assert_allclose(wnr.predict(X), y)

    y_pred = wnr.predict(X_test)
    euclidean = np.sqrt(((X_test[:, np.newaxis, :] - X) ** 2).sum(axis=-1))
    w = 1 / euclidean
    w /= w.sum(axis=1, keepdims=True)

    assert_allclose(y_pred, (w @ y))


def test_weighted_neighbors_precomputed():
    X, y = make_regression(n_samples=5000, n_features=5000, n_targets=100)
    X = np.abs(X)
    X.setflags(write=False)

    wnr = WeightedNeighborsRegressor(metric="precomputed")
    wnr.fit(X, y)
    wnr.predict(X[:X.shape[0]//2])

    y_pred = wnr.predict(X)

    w = (1/X)
    w /= w.sum(axis=1, keepdims=True)
    assert (y_pred == w @ y).all()


def main():
    test_weighted_neighbors_euclidean()
    test_weighted_neighbors_precomputed()


if __name__ == '__main__':
    main()


"""
The :mod:`sklearn.ensemble` module includes ensemble-based methods for
classification, regression and anomaly detection.
"""
from ._forest import (
    BipartiteRandomForestRegressor,
    BipartiteExtraTreesRegressor,
)
from ._semisupervised_forest import *
from ._gb import *

__all__ = [
    "BipartiteRandomForestRegressor"
    "BipartiteExtraTreesRegressor",
]



# cython: boundscheck=False
#
# Since splitter objects are not exported from sklearn.tree._splitter,
# this module must exist to introduce a few lines to BestSplitter.
# 
# Author: Pedro Ilidio
# Adapted from https://github.com/scikit-learn/scikit-learn/blob/1.1.1/sklearn/tree/_splitter.pyx
#
# License: BSD 3 clause

import numpy as np

from libc.stdlib cimport free
from libc.stdlib cimport qsort
from libc.string cimport memcpy
from libc.string cimport memset

from sklearn.tree._splitter cimport (
    Splitter, DTYPE_t, SIZE_t, DOUBLE_t, UINT32_t, SplitRecord,
)
from sklearn.tree._utils cimport rand_int, rand_uniform, log
from sklearn.tree._criterion cimport Criterion


cdef double INFINITY = np.inf
# Mitigate precision differences between 32 bit and 64 bit
cdef DTYPE_t FEATURE_THRESHOLD = 1e-7


cdef inline void _init_split(SplitRecord* self, SIZE_t start_pos) nogil:
    self.impurity_left = INFINITY
    self.impurity_right = INFINITY
    self.pos = start_pos
    self.feature = 0
    self.threshold = 0.
    self.improvement = -INFINITY


cdef class BaseDenseSplitter(Splitter):
    cdef const DTYPE_t[:, :] X

    cdef SIZE_t n_total_samples

    cdef int init(self,
                  object X,
                  const DOUBLE_t[:, ::1] y,
                  DOUBLE_t[:] sample_weight) except -1:
        """Initialize the splitter
        Returns -1 in case of failure to allocate memory (and raise MemoryError)
        or 0 otherwise.
        """

        # Call parent init
        Splitter.init(self, X, y, sample_weight)

        self.X = X
        return 0


cdef class BestSplitterSFSS(BaseDenseSplitter):
    """Splitter for finding the best split."""
    def __reduce__(self):
        return (BestSplitterSFSS, (self.criterion,
                               self.max_features,
                               self.min_samples_leaf,
                               self.min_weight_leaf,
                               self.random_state), self.__getstate__())

    cdef int node_split(self, double impurity, SplitRecord* split,
                        SIZE_t* n_constant_features) nogil except -1:
        """Find the best split on node samples[start:end]
        Returns -1 in case of failure to allocate memory (and raise MemoryError)
        or 0 otherwise.
        """
        # Find the best split
        cdef SIZE_t[::1] samples = self.samples
        cdef SIZE_t start = self.start
        cdef SIZE_t end = self.end

        cdef SIZE_t[::1] features = self.features
        cdef SIZE_t[::1] constant_features = self.constant_features
        cdef SIZE_t n_features = self.n_features

        cdef DTYPE_t[::1] Xf = self.feature_values
        cdef SIZE_t max_features = self.max_features
        cdef SIZE_t min_samples_leaf = self.min_samples_leaf
        cdef double min_weight_leaf = self.min_weight_leaf
        cdef UINT32_t* random_state = &self.rand_r_state

        cdef SplitRecord best, current
        cdef double current_proxy_improvement = -INFINITY
        cdef double best_proxy_improvement = -INFINITY

        cdef SIZE_t f_i = n_features
        cdef SIZE_t f_j
        cdef SIZE_t p
        cdef SIZE_t feature_idx_offset
        cdef SIZE_t feature_offset
        cdef SIZE_t i
        cdef SIZE_t j

        cdef SIZE_t n_visited_features = 0
        # Number of features discovered to be constant during the split search
        cdef SIZE_t n_found_constants = 0
        # Number of features known to be constant and drawn without replacement
        cdef SIZE_t n_drawn_constants = 0
        cdef SIZE_t n_known_constants = n_constant_features[0]
        # n_total_constants = n_known_constants + n_found_constants
        cdef SIZE_t n_total_constants = n_known_constants
        cdef DTYPE_t current_feature_value
        cdef SIZE_t partition_end

        cdef double _wnns

        _init_split(&best, end)

        # Sample up to max_features without replacement using a
        # Fisher-Yates-based algorithm (using the local variables `f_i` and
        # `f_j` to compute a permutation of the `features` array).
        #
        # Skip the CPU intensive evaluation of the impurity criterion for
        # features that were already detected as constant (hence not suitable
        # for good splitting) by ancestor nodes and save the information on
        # newly discovered constant features to spare computation on descendant
        # nodes.
        while (f_i > n_total_constants and  # Stop early if remaining features
                                            # are constant
                (n_visited_features < max_features or
                 # At least one drawn features must be non constant
                 n_visited_features <= n_found_constants + n_drawn_constants)):

            n_visited_features += 1

            # Loop invariant: elements of features in
            # - [:n_drawn_constant[ holds drawn and known constant features;
            # - [n_drawn_constant:n_known_constant[ holds known constant
            #   features that haven't been drawn yet;
            # - [n_known_constant:n_total_constant[ holds newly found constant
            #   features;
            # - [n_total_constant:f_i[ holds features that haven't been drawn
            #   yet and aren't constant apriori.
            # - [f_i:n_features[ holds features that have been drawn
            #   and aren't constant.

            # Draw a feature at random
            f_j = rand_int(n_drawn_constants, f_i - n_found_constants,
                           random_state)

            if f_j < n_known_constants:
                # f_j in the interval [n_drawn_constants, n_known_constants[
                features[n_drawn_constants], features[f_j] = features[f_j], features[n_drawn_constants]

                n_drawn_constants += 1

            else:
                # f_j in the interval [n_known_constants, f_i - n_found_constants[
                f_j += n_found_constants
                # f_j in the interval [n_total_constants, f_i[
                current.feature = features[f_j]

              # ===============================================
              # bipartite_learn NOTE:                              

                # Sort samples along that feature; by
                # copying the values into an array and
                # sorting the array in a manner which utilizes the cache more
                # effectively.
                for i in range(start, end):
                    Xf[i] = self.X[samples[i], current.feature]

                sort(&Xf[start], &samples[start], end - start)

                if Xf[end - 1] <= Xf[start] + FEATURE_THRESHOLD:
                    features[f_j], features[n_total_constants] = features[n_total_constants], features[f_j]

                    n_found_constants += 1
                    n_total_constants += 1

                else:
                    f_i -= 1
                    features[f_i], features[f_j] = features[f_j], features[f_i]

                # ===============================================
                # One of the two lines that we added.
                # ===============================================
                # TODO: Release GIL.                            
                # ===============================================
                    with gil:
                        self.criterion.set_feature(current.feature)
                # ===============================================

                    # Evaluate all splits
                    self.criterion.reset()
                    p = start

                    while p < end:
                        while (p + 1 < end and
                               Xf[p + 1] <= Xf[p] + FEATURE_THRESHOLD):
                            p += 1

                        # (p + 1 >= end) or (X[samples[p + 1], current.feature] >
                        #                    X[samples[p], current.feature])
                        p += 1
                        # (p >= end) or (X[samples[p], current.feature] >
                        #                X[samples[p - 1], current.feature])

                        if p < end:
                            current.pos = p

                            # Reject if min_samples_leaf is not guaranteed
                            if (((current.pos - start) < min_samples_leaf) or
                                    ((end - current.pos) < min_samples_leaf)):
                                continue

                            self.criterion.update(current.pos)

                            # Reject if min_weight_leaf is not satisfied
                            if ((self.criterion.weighted_n_left < min_weight_leaf) or
                                    (self.criterion.weighted_n_right < min_weight_leaf)):
                                continue

                            current_proxy_improvement = self.criterion.proxy_impurity_improvement()

                            if current_proxy_improvement > best_proxy_improvement:
                                best_proxy_improvement = current_proxy_improvement
                                # sum of halves is used to avoid infinite value
                                current.threshold = Xf[p - 1] / 2.0 + Xf[p] / 2.0

                                if ((current.threshold == Xf[p]) or
                                    (current.threshold == INFINITY) or
                                    (current.threshold == -INFINITY)):
                                    current.threshold = Xf[p - 1]

                                best = current  # copy

        # Reorganize into samples[start:best.pos] + samples[best.pos:end]
        if best.pos < end:
            partition_end = end
            p = start

            while p < partition_end:
                if self.X[samples[p], best.feature] <= best.threshold:
                    p += 1

                else:
                    partition_end -= 1

                    samples[p], samples[partition_end] = samples[partition_end], samples[p]

          # ============================================
          # bipartite_learn NOTE:
          # ============================================
            with gil:
                self.criterion.set_feature(best.feature)
          # ============================================

            self.criterion.reset()
            self.criterion.update(best.pos)
            self.criterion.children_impurity(&best.impurity_left,
                                             &best.impurity_right)
            best.improvement = self.criterion.impurity_improvement(
                impurity, best.impurity_left, best.impurity_right)

        # Respect invariant for constant features: the original order of
        # element in features[:n_known_constants] must be preserved for sibling
        # and child nodes
        memcpy(&features[0], &constant_features[0], sizeof(SIZE_t) * n_known_constants)

        # Copy newly found constant features
        memcpy(&constant_features[n_known_constants],
               &features[n_known_constants],
               sizeof(SIZE_t) * n_found_constants)

        # Return values
        split[0] = best
        n_constant_features[0] = n_total_constants
        return 0


# Sort n-element arrays pointed to by Xf and samples, simultaneously,
# by the values in Xf. Algorithm: Introsort (Musser, SP&E, 1997).
cdef inline void sort(DTYPE_t* Xf, SIZE_t* samples, SIZE_t n) nogil:
    if n == 0:
      return
    cdef int maxd = 2 * <int>log(n)
    introsort(Xf, samples, n, maxd)


cdef inline void swap(DTYPE_t* Xf, SIZE_t* samples,
        SIZE_t i, SIZE_t j) nogil:
    # Helper for sort
    Xf[i], Xf[j] = Xf[j], Xf[i]
    samples[i], samples[j] = samples[j], samples[i]


cdef inline DTYPE_t median3(DTYPE_t* Xf, SIZE_t n) nogil:
    # Median of three pivot selection, after Bentley and McIlroy (1993).
    # Engineering a sort function. SP&E. Requires 8/3 comparisons on average.
    cdef DTYPE_t a = Xf[0], b = Xf[n // 2], c = Xf[n - 1]
    if a < b:
        if b < c:
            return b
        elif a < c:
            return c
        else:
            return a
    elif b < c:
        if a < c:
            return a
        else:
            return c
    else:
        return b


# Introsort with median of 3 pivot selection and 3-way partition function
# (robust to repeated elements, e.g. lots of zero features).
cdef void introsort(DTYPE_t* Xf, SIZE_t *samples,
                    SIZE_t n, int maxd) nogil:
    cdef DTYPE_t pivot
    cdef SIZE_t i, l, r

    while n > 1:
        if maxd <= 0:   # max depth limit exceeded ("gone quadratic")
            heapsort(Xf, samples, n)
            return
        maxd -= 1

        pivot = median3(Xf, n)

        # Three-way partition.
        i = l = 0
        r = n
        while i < r:
            if Xf[i] < pivot:
                swap(Xf, samples, i, l)
                i += 1
                l += 1
            elif Xf[i] > pivot:
                r -= 1
                swap(Xf, samples, i, r)
            else:
                i += 1

        introsort(Xf, samples, l, maxd)
        Xf += r
        samples += r
        n -= r


cdef inline void sift_down(DTYPE_t* Xf, SIZE_t* samples,
                           SIZE_t start, SIZE_t end) nogil:
    # Restore heap order in Xf[start:end] by moving the max element to start.
    cdef SIZE_t child, maxind, root

    root = start
    while True:
        child = root * 2 + 1

        # find max of root, left child, right child
        maxind = root
        if child < end and Xf[maxind] < Xf[child]:
            maxind = child
        if child + 1 < end and Xf[maxind] < Xf[child + 1]:
            maxind = child + 1

        if maxind == root:
            break
        else:
            swap(Xf, samples, root, maxind)
            root = maxind


cdef void heapsort(DTYPE_t* Xf, SIZE_t* samples, SIZE_t n) nogil:
    cdef SIZE_t start, end

    # heapify
    start = (n - 2) // 2
    end = n
    while True:
        sift_down(Xf, samples, start, end)
        if start == 0:
            break
        start -= 1

    # sort by shrinking the heap, putting the max element immediately after it
    end = n - 1
    while end > 0:
        swap(Xf, samples, 0, end)
        sift_down(Xf, samples, 0, end)
        end = end - 1



cdef class RandomSplitterSFSS(BaseDenseSplitter):
    """Splitter for finding the best random split."""
    def __reduce__(self):
        return (RandomSplitterSFSS, (self.criterion,
                                     self.max_features,
                                     self.min_samples_leaf,
                                     self.min_weight_leaf,
                                     self.random_state), self.__getstate__())

    cdef int node_split(self, double impurity, SplitRecord* split,
                        SIZE_t* n_constant_features) nogil except -1:
        """Find the best random split on node samples[start:end]
        Returns -1 in case of failure to allocate memory (and raise MemoryError)
        or 0 otherwise.
        """
        # Draw random splits and pick the best
        cdef SIZE_t[::1] samples = self.samples
        cdef SIZE_t start = self.start
        cdef SIZE_t end = self.end

        cdef SIZE_t[::1] features = self.features
        cdef SIZE_t[::1] constant_features = self.constant_features
        cdef SIZE_t n_features = self.n_features

        cdef DTYPE_t[::1] Xf = self.feature_values
        cdef SIZE_t max_features = self.max_features
        cdef SIZE_t min_samples_leaf = self.min_samples_leaf
        cdef double min_weight_leaf = self.min_weight_leaf
        cdef UINT32_t* random_state = &self.rand_r_state

        cdef SplitRecord best, current
        cdef double current_proxy_improvement = - INFINITY
        cdef double best_proxy_improvement = - INFINITY

        cdef SIZE_t f_i = n_features
        cdef SIZE_t f_j
        cdef SIZE_t p
        cdef SIZE_t partition_end
        cdef SIZE_t feature_stride
        # Number of features discovered to be constant during the split search
        cdef SIZE_t n_found_constants = 0
        # Number of features known to be constant and drawn without replacement
        cdef SIZE_t n_drawn_constants = 0
        cdef SIZE_t n_known_constants = n_constant_features[0]
        # n_total_constants = n_known_constants + n_found_constants
        cdef SIZE_t n_total_constants = n_known_constants
        cdef SIZE_t n_visited_features = 0
        cdef DTYPE_t min_feature_value
        cdef DTYPE_t max_feature_value
        cdef DTYPE_t current_feature_value

        _init_split(&best, end)

        # Sample up to max_features without replacement using a
        # Fisher-Yates-based algorithm (using the local variables `f_i` and
        # `f_j` to compute a permutation of the `features` array).
        #
        # Skip the CPU intensive evaluation of the impurity criterion for
        # features that were already detected as constant (hence not suitable
        # for good splitting) by ancestor nodes and save the information on
        # newly discovered constant features to spare computation on descendant
        # nodes.
        while (f_i > n_total_constants and  # Stop early if remaining features
                                            # are constant
                (n_visited_features < max_features or
                 # At least one drawn features must be non constant
                 n_visited_features <= n_found_constants + n_drawn_constants)):
            n_visited_features += 1

            # Loop invariant: elements of features in
            # - [:n_drawn_constant[ holds drawn and known constant features;
            # - [n_drawn_constant:n_known_constant[ holds known constant
            #   features that haven't been drawn yet;
            # - [n_known_constant:n_total_constant[ holds newly found constant
            #   features;
            # - [n_total_constant:f_i[ holds features that haven't been drawn
            #   yet and aren't constant apriori.
            # - [f_i:n_features[ holds features that have been drawn
            #   and aren't constant.

            # Draw a feature at random
            f_j = rand_int(n_drawn_constants, f_i - n_found_constants,
                           random_state)

            if f_j < n_known_constants:
                # f_j in the interval [n_drawn_constants, n_known_constants[
                features[n_drawn_constants], features[f_j] = features[f_j], features[n_drawn_constants]
                n_drawn_constants += 1

            else:
                # f_j in the interval [n_known_constants, f_i - n_found_constants[
                f_j += n_found_constants
                # f_j in the interval [n_total_constants, f_i[

                current.feature = features[f_j]

                # Find min, max
                min_feature_value = self.X[samples[start], current.feature]
                max_feature_value = min_feature_value
                Xf[start] = min_feature_value

                for p in range(start + 1, end):
                    current_feature_value = self.X[samples[p], current.feature]
                    Xf[p] = current_feature_value

                    if current_feature_value < min_feature_value:
                        min_feature_value = current_feature_value
                    elif current_feature_value > max_feature_value:
                        max_feature_value = current_feature_value

                if max_feature_value <= min_feature_value + FEATURE_THRESHOLD:
                    features[f_j], features[n_total_constants] = features[n_total_constants], current.feature

                    n_found_constants += 1
                    n_total_constants += 1

                else:
                    f_i -= 1
                    features[f_i], features[f_j] = features[f_j], features[f_i]

                    # Draw a random threshold
                    current.threshold = rand_uniform(min_feature_value,
                                                     max_feature_value,
                                                     random_state)

                    if current.threshold == max_feature_value:
                        current.threshold = min_feature_value

                    # Partition
                    p, partition_end = start, end
                    while p < partition_end:
                        if Xf[p] <= current.threshold:
                            p += 1
                        else:
                            partition_end -= 1

                            Xf[p], Xf[partition_end] = Xf[partition_end], Xf[p]
                            samples[p], samples[partition_end] = samples[partition_end], samples[p]

                    current.pos = partition_end

                    # Reject if min_samples_leaf is not guaranteed
                    if (((current.pos - start) < min_samples_leaf) or
                            ((end - current.pos) < min_samples_leaf)):
                        continue

                # ===============================================
                # bipartite_learn NOTE:                              
                # ===============================================
                # One of the two lines that we added.
                # ===============================================
                # TODO: Release GIL.                            
                # ===============================================
                    with gil:
                        self.criterion.set_feature(current.feature)
                # ===============================================

                    # Evaluate split
                    self.criterion.reset()
                    self.criterion.update(current.pos)

                    # Reject if min_weight_leaf is not satisfied
                    if ((self.criterion.weighted_n_left < min_weight_leaf) or
                            (self.criterion.weighted_n_right < min_weight_leaf)):
                        continue

                    current_proxy_improvement = self.criterion.proxy_impurity_improvement()

                    if current_proxy_improvement > best_proxy_improvement:
                        best_proxy_improvement = current_proxy_improvement
                        best = current  # copy

        # Reorganize into samples[start:best.pos] + samples[best.pos:end]
        if best.pos < end:
            if current.feature != best.feature:
                p, partition_end = start, end

                while p < partition_end:
                    if self.X[samples[p], best.feature] <= best.threshold:
                        p += 1
                    else:
                        partition_end -= 1

                        samples[p], samples[partition_end] = samples[partition_end], samples[p]

        # ===============================================
        # bipartite_learn NOTE:                              
        # ===============================================
        # One of the two lines that we added.
        # ===============================================
        # TODO: Release GIL.                            
        # ===============================================
            with gil:
                self.criterion.set_feature(current.feature)
        # ===============================================

            self.criterion.reset()
            self.criterion.update(best.pos)
            self.criterion.children_impurity(&best.impurity_left,
                                             &best.impurity_right)
            best.improvement = self.criterion.impurity_improvement(
                impurity, best.impurity_left, best.impurity_right)

        # Respect invariant for constant features: the original order of
        # element in features[:n_known_constants] must be preserved for sibling
        # and child nodes
        memcpy(&features[0], &constant_features[0], sizeof(SIZE_t) * n_known_constants)

        # Copy newly found constant features
        memcpy(&constant_features[n_known_constants],
               &features[n_known_constants],
               sizeof(SIZE_t) * n_found_constants)

        # Return values
        split[0] = best
        n_constant_features[0] = n_total_constants
        return 0


import joblib
import pytest
from pathlib import Path
from sklearn.tree import DecisionTreeRegressor, export_text
from sklearn.ensemble import RandomForestRegressor
from sklearn.base import clone
from sklearn.model_selection._validation import _score
from sklearn.metrics import check_scoring
from bipartite_learn.wrappers import GlobalSingleOutputWrapper
from bipartite_learn.melter import melt_multipartite_dataset
from .utils.make_examples import make_interaction_regression
from .utils.test_utils import stopwatch

with stopwatch('Making dataset'):
    XX, Y, X, y = make_interaction_regression(
        return_molten=True,
        n_samples=(50, 60),
        n_features=(10, 9),
        random_state=0,
        max_target=1.0,
    )
    # Make it binary  # TODO: specialized data generation function
    y_mean = Y.mean()
    Y = (Y > y_mean).astype(int)
    y = (y > y_mean).astype(int)
    # Ensure negative majority
    if Y.mean() > 0.5:
        Y, y = 1-Y, 1-y


@pytest.fixture(params=range(10))
def random_state(request):
    return request.param


@pytest.fixture(params=[1, 0.1])
def min_samples_leaf(request):
    return request.param


def test_pu_wrapper_bipartite_no_neg_subsamples(random_state, min_samples_leaf):
    Y_ = (Y == 1).astype(int)
    y_ = (y == 1).astype(int)

    tree = DecisionTreeRegressor(
        min_samples_leaf=min_samples_leaf,
        random_state=random_state,
    )
    tree_bipartite = GlobalSingleOutputWrapper(clone(tree), subsample_negatives=False)

    tree_bipartite.fit(XX, Y_)
    tree.fit(X, y_)

    assert export_text(tree_bipartite.estimator) == export_text(tree)


def test_pu_wrapper_bipartite_neg_subsamples(random_state, min_samples_leaf):
    tree = DecisionTreeRegressor(
        min_samples_leaf=min_samples_leaf,
        random_state=random_state,
    )
    tree_bipartite = GlobalSingleOutputWrapper(
        estimator=clone(tree),
        subsample_negatives=True,
        random_state=random_state,
    )
    Xs, ys = melt_multipartite_dataset(
        XX, Y,
        subsample_negatives=True,
        random_state=random_state,
    )

    print(f"Keeping {ys.shape[0]} out of {y.shape[0]}. "
          f"Relative new size: {ys.shape[0]/y.shape[0]}\n")

    assert ys.mean() == .5
    assert ys.shape[0]/y.shape[0] == 2*y.mean()
    assert ys.shape[0] < y.shape[0]
    assert Xs.shape[0] == ys.shape[0]

    tree_bipartite.fit(XX, Y)
    tree.fit(Xs, ys)

    # print(export_text(tree_bipartite.estimator))
    # print(export_text(tree))

    assert export_text(tree_bipartite.estimator) == export_text(tree)


def _test_pickling(obj):
    joblib.dump(obj, 'test.pickle')
    obj_loaded = joblib.load('test.pickle')
    Path('test.pickle').unlink()

    obj.get_params()
    obj_loaded.get_params()


def test_wrapped_tree_pickling():
    tree = DecisionTreeRegressor(
        min_samples_leaf=30,
        random_state=0,
    )
    tree_bipartite = GlobalSingleOutputWrapper(clone(tree), subsample_negatives=True)

    _test_pickling(tree)
    _test_pickling(tree_bipartite)


def test_wrapped_forest_pickling():
    forest = RandomForestRegressor(
        min_samples_leaf=30,
        random_state=0,
    )
    forest_bipartite = GlobalSingleOutputWrapper(
        clone(forest), subsample_negatives=True)

    _test_pickling(forest)
    _test_pickling(forest_bipartite)


def test_score():
    forest = RandomForestRegressor(
        min_samples_leaf=30,
        random_state=0,
    )
    forest_bipartite = GlobalSingleOutputWrapper(
        clone(forest),
        subsample_negatives=True,
        random_state=0,
    )
    forest_bipartite.fit(XX, Y)

    score = _score(
        forest_bipartite, X, y,
        check_scoring(forest_bipartite, "roc_auc"),
        error_score="raise",
    )
    print("Score:", score)



import numpy as np
from sklearn.utils.validation import check_random_state
from sklearn.utils._param_validation import validate_params
from bipartite_learn.base import (
    BaseBipartiteEstimator,
    BaseMultipartiteSampler,
)
from bipartite_learn.utils import _X_is_multipartite

__all__ = [
    "BipartiteMelter",
    "MultiPartiteMelter",
    "melt_multipartite_dataset",
]


class BipartiteMelter(BaseMultipartiteSampler, BaseBipartiteEstimator):
    """Convert a bipartite GMO database to a simpler GSO database.

    Convert a bipartite interaction problem, where there are two feature
    matrices in X (one for each axis) and an interaction matrix y to a simpler
    usual format where each sample is a combination of samples from X[0] and
    X[1].

    Slightly faster than MultipartiteMelter
    """

    def _fit_resample(self, X, y):
        X1, X2 = X
        melted_X = np.hstack([
            np.repeat(X1, X2.shape[0], axis=0),
            np.tile(X2, (X1.shape[0], 1))
        ])
        melted_y = y.reshape(-1)
        return melted_X, melted_y


class MultipartiteMelter(BaseMultipartiteSampler):
    """Convert a n-dimensional GMO database to a simpler GSO database."""

    def __init__(self, subsample_negatives=False, random_state=None):
        self.subsample_negatives = subsample_negatives
        self.random_state = random_state

    def _fit_resample(self, X, y=None):
        return melt_multipartite_dataset(
            X,
            y,
            subsample_negatives=self.subsample_negatives,
            # Delegating check_random_state since samplers are stateless
            random_state=self.random_state,
        )


def row_cartesian_product(X):
    """Row cartesian product of 2D arrays in X.

    Pick one row from each of the 2D arrays in X, in their presented order, and
    concatenate them. Repeat. Return a 2D array where its rows are all the
    possible combinations of rows in X.

    Parameters
    ----------
    X : list-like of 2D np.ndarrays

    Returns
    -------
    result : 2D np.ndarray
        Cartesian product of X's 2d arrays, row-wise.
    """
    lengths = [Xi.shape[0] for Xi in X]
    # Xi will repeat once for every combination of samples ahead (Xj if j>i)
    # and tile once for every combination of samples before (Xj if j<i).
    n_repeat_tile = [
        (np.prod(lengths[(i+1):]), np.prod(lengths[:i]))
        for i in range(len(X))
    ]

    result = np.hstack([
        np.tile(Xi.repeat(n_repeat, axis=0), (n_tile, 1))
        for Xi, (n_repeat, n_tile) in zip(X, n_repeat_tile)
    ])

    return result


@validate_params({
    'X': ['array-like'],
    'y': ['array-like', None],
    'subsample_negatives': ['boolean'],
    'random_state': ['random_state'],
})
def melt_multipartite_dataset(
    X, y=None, subsample_negatives=False, random_state=None,
):
    """Melt bipartite input.
    
    If X is a list of Xi feature matrices, one for each bipartite group,
    convert it to traditional data format by generating concatenations of
    rows from X[0] with rows from X[1].
    """
    if not _X_is_multipartite(X):
        raise ValueError(
            "Tried to melt monopartite dataset. X must be a list/tuple of "
            "array-likes."
        )
    X = row_cartesian_product(X) 
    if y is None:
        return X, y

    y = y.reshape(-1)

    if subsample_negatives:
        if set(np.unique(y)) != {0, 1}:
            raise ValueError(
                "y must have only 0 or 1 elements if subsample_negatives=True."
            )

        random_state = check_random_state(random_state)
        pos_negatives = np.nonzero(1 - y)[0]
        # difference in number of positives vs number of negatives
        # i. e. number of negatives minus number of positives
        delta = 2 * pos_negatives.size - y.size

        if delta < 0:
            raise ValueError(
                "There must be less positive than negative values to subsample"
                "negatives."
            )

        # Subsample negatives to have the same size as positives
        negatives_to_del = random_state.choice(
            pos_negatives,
            replace=False,
            size=delta,
        )
        X = np.delete(X, negatives_to_del, axis=0)
        y = np.delete(y, negatives_to_del, axis=0)
    
    return X, y

